// file: fms.cc

// enthaelt Netzobjekt

#include <iostream.h>
#include <fstream.h>
#include <new.h>
#include <math.h>
#include "io.h" // wg. io_move
#include "general.h" // muss vor ms.h stehen wegen boolean!
#include "fms.h"
#include "mweights.h"
#include "patterns.h"

//
//  Objekt Netz
//
class Netz {
public:
  double  *netinput;
  double  *target;
  double  *u_error;
  int av_sigs, min_sigs, max_sigs;
  
  //fms
  boolean **alive, **wasalive, **insignificant;
  double E_gamma, epsilon, **t_eins, *t_zwei, t_drei, t_vier;
  double del_min;
  double **ks, ***ykw, **yw;
  double **ky, ***b, **del_w;
  double **ry, **rs, **rdw, *rdky, *rdks;
  long exemplars;
  int sig_weights; // Anzahl der signifikanten Gewichte
  double lyw, lrdw, scale;
  double B, B_a, del_max;
  int are_alive, alive_units; // Anzahl der Gewichte mit alive[][]==TRUE
  double *un_av, *un_min, *un_max, *un_var;// durchschnittl. Akt. einer Unit  
  int * un_upper, *un_lower; // Kriterien fuer Binaeritaet
  boolean *treated; // zum update von alive: ob unit schon behandelt worden ist
  boolean init_error; // zur Initialisierung von B_a und E_a in der ersten Epoche

  Netz();                   // Initialisierung: Netzdefinitionsdatei lesen, globale und Netzvariablen initialisieren
  void Run();               // Trainingslauf mit Bildschirmausgabe
  ~Netz();                  // Variablen freigeben, Gewichte sichern, ...
  double log_output (double x); //Aktivierungsfunktion
  double log_hidden (double x); //Aktivierungsfunktion
  double log_out_strich (double x); // erste Ableitung
  double log_hid_strich (double x); // erste Ableitung
  double log_out_2strich (double x); // zweite Ableitung
  double log_hid_2strich (double x); // zweite Ableitung
  void compute_output();    // Netzoutput berechnen
  void init_params();      // Parameter initialisieren
  void init_weights();      // Gewichte initialisieren
  void compute_error();     // Fehler berechnen
  void compute_wed();       // Gewichtsfehlerableitungen berechnen
  void compute_t2_t3();     // t2 und t3 berechnen 
  void compute_insig_weights();   // insignifikante Gewichte berechnen 
  void update_variables();  // Variablen-Update nach compute_insig_weights()
  void compute_t4();        // t4 und b[][][] ausrechnen
  void sec_forward();       // zweiter forward-pass
  void sec_backward();      // zweiter backward-pass
  void norm_grad_b();       // Normalisierung des Gradienten von B
  void change_weights();    // Gewichte anpassen
  void setinput();          // Muster an Eingabeunits anlegen
  void settarget();         // Referenzmuster setzen
  void setup_pattern();     // setinput() und settarget()
  int trial();
  int train(char c);        // Netz trainieren
  int ptrain();             // random-Training
  int strain();             // serielles Training
  int test_all();           // alle Muster testen

  void update_lparam();     // passt die Lernparameter nach lparam_update - Epochen an
  int write_activations();  // schreibt aktuelle Aktivierungen, Minima und Maxima in die Datei xxxxxxxx.act
  int update_display();     // vorher in display.cc zum Anzeigen der aktuellen Parameter
  int write_alives();       // schreibt aktuelle alive[][], wasalive[][] xxxxxxxx.alive
  int read_alives();        // liest aktuelle alive[][], wasalive[][] xxxxxxxx.alive
  int write_netinputs();
  int write_unit_acts();   // un_av, un_min, un_max, un_lower, un_upper
  int comp_alive_units();  // zaehlt alle Units, die noch nicht weggeloescht wurden
};
// Netz Ende
static void heap_full()
{
  cerr << "\n\aDer Heap ist voll.";
  exit(1);
}

// Globale Variablen ******************************
extern int nunits; //in weights.h
extern int ninputs;
extern int noutputs;
double tss; // summierter quadrierter Fehler
double pss_max; // maximaler quadrierter Fehler
double *activation; // wg. display
double ecrit;
boolean System_Defined = FALSE;
int *num_weights_to;
char *Default_step_string = "epoche";
char grain_string[20] = "pattern";
char choose_string[20] = "random";
boolean lflag=TRUE;
boolean statlog=FALSE;
int nepochs;
int patno; // gerade bearbeitetes Muster
// Globale Variablen Ende ***********************

// Variablen fuer FMS ****************************************************
double sigma; // Aenderung von lambda in Spezialfall E_n > E_a
double delta_lambda;
double E_a = 0; // geglaetteter Fehler: aus E_n
double E_n = 0; // ungeglaetteter Fehler: tss/npatterns
double E_o = 0; // alter Fehler E_n
double E = 0; // quadratischer Fehler E des aktuellen Musters
int lparam_update=1; // Anzahl Epochen nach denen die Lernparameter angepasst werden
int tpatterns=1; // Anzahl der fuer eine Epoche verwendeten Muster
double up_border=.9; // von target abhaengig
double low_border=.1; // s.o.
int correct_class=0;// Anzahl der pro Epoche korrekt erkannten Muster
//  **********************************************************************


int put_error(char *errorstring)
{
  cout << "\n" << errorstring;
  return(0);
}

extern int read_weights();
extern int write_weights();

boolean start(char *s1, char *s2)   
// faengt s1 genauso an wie s2?
{
  while (*s1 && *s2) {
    if (*s1++ != *s2++)
      return(FALSE);
  }
  if(*s1 && !*s2)  // wenn s1 laenger ist als s2 soll die Fkt. FALSE liefern
    return(FALSE);
  return(TRUE);
}

Netz::Netz() {
  int i;
    
// public-Variablen initialisieren *******************
  System_Defined = FALSE;
  nunits = 0;
  ninputs = 0;
  noutputs = 0;
  lflag = TRUE; // standardmaessig lernen!!!
  epochno = 0;
  nepochs = 100;
  patno = 0;
  tss = 0.0;
  pss_max = 0.0;
  ecrit = 0.0;
  netinput = NULL;
  activation = NULL;
  u_error = NULL;
  target = NULL;

  // FMS
  alpha=.01;
  sigma = .5; // Aenderung von lambda
  delta_lambda = .1*alpha;
  lambda = 0; // vorlaeufige Initialisierung mit kleinem Wert
  K=pow(10, 3);
  epsilon=0.01;
  exemplars=epochno=0;
  E_gamma=.9;
  ecrit=.1;
  tss=E_a=E_n=E_o=0;
  B=0;
  B_a=0;
  init_error=FALSE;

  // *************************************************************
  // Netz-Definitionsdatei einlesen, falls vorhanden
  #ifdef DEBUG
  cout << "\nfms|netz::define_bp_network()";
  #endif
  define_bp_network(); // liest globale Variablen und Netzwerkbeschreibung ein.
                       // in mweights.cc implementiert
  // Ende globale Variablen
  if (!nunits) {
    put_error("BP-Netz kann nicht initialisiert werden - nunits nicht definiert!");
    return;
  }
  else
  if (!noutputs) {
    put_error("BP-Netz kann nicht initialisiert werden - noutputs nicht definiert!");
    return;
  }
  else
  if (!ninputs) {
    put_error("BP-Netz kann nicht initialisiert werden - ninputs nicht definiert!");
    return;
  }
  else
  if (!(nunits && noutputs && ninputs)) {
    put_error("BP-Netz kann nicht initialisiert werden - nunits - noutputs - ninputs nicht definiert!");
    return;
  }

  netinput = new double[nunits];
  for (i = 0; i < nunits; i++)
    netinput[i] = 0.0;

  activation = new double[nunits];
  for (i = 0; i < nunits; i++)
    activation[i] = 0.5;

  u_error = new double[nunits];
  for (i = 0; i < nunits; i++)
    u_error[i] = 0.0;

  target = new double[noutputs];
  for (i = 0; i < noutputs; i++)
    target[i] = 0.0;

  System_Defined = TRUE;

  // Gewichte initialisieren
  init_weights();

  // Trainings-Muster einlesen aus Datei "NETZ.MST", ist in patterns.cpp
  init_pattern_pairs();
  init_params();
}

Netz::~Netz(){
  #ifdef DEBUG
  cout << "\nfms|netz::~netz()";
  #endif
  // Gewichte schreiben
  if (lflag)
    write_weights();
}

void Netz::Run(){
  #ifdef DEBUG
  cout << "\nfms|netz::Run";
  #endif
  
  // nach Art des Aufrufs: lernen oder testen mit den eingestellten Werten
  if (lflag) {
  // Lernmodus
    if (start(choose_string, "random")) {
    // zufaellige Musterauswahl
      ptrain();
    }
    else
    // serielle Musterauswahl
      strain(); 
  }
  else {
  // Testmodus lflag==0
  // testet alle Muster der angegebenen Datei
    test_all();
  }
}

double  Netz::log_hidden(double x)
{
  return((1.0 / (1.0 + (double) exp( (double) ((-1.0) * x)))));
}

double  Netz::log_output(double x)
{
  return((2.0 / (1.0 + (double) exp( (double) (-1.0 * x))))-1.0);
}

double Netz::log_out_strich(double x)
{
  double temp;
  // erste Ableitung von logistic
  temp=log_output(x);
  return(.5*((double)(temp+1.0))*(double)(1.0-temp));
}

double Netz::log_out_2strich(double x)
{
  double temp;
  // zweite Ableitung von logistic
  temp = log_output(x);
  return(-.5*(double)temp*((double)temp+1.0)*(1.0-(double)temp));
}

double Netz::log_hid_strich(double x)
{
  double temp;
  // erste Ableitung von logistic
  temp=log_hidden(x);
  return((double)temp*(1.0-(double)temp));
}

double Netz::log_hid_2strich(double x)
{
  double temp;
  // zweite Ableitung von logistic
  temp = log_hidden(x);
  return((double)temp*(1.0-(double)temp)*(1.0-2.0*(double)temp));
}

void Netz::compute_output() {
  register int i, j;
  double net;
  int grenze1, grenze2;
  
  #ifdef DEBUG
  cout << "\nfms|Netz::compute_output()";
  #endif
  for (i = ninputs; i < nunits; i++) {// hidden-output
    net = 0;
      
    if (i>=nunits-noutputs) // i in hidden schicht
    {
      grenze1=ninputs;
      grenze2=nunits-noutputs;
    }
    else// i in output schicht
    {
      grenze1=0;
      grenze2=ninputs;
    }
				      
    for (j=grenze1; j<grenze2; j++)// input-hidden
    {
      if (alive[i][j]) 
        net += (activation[j])*(weight[i][j]);
    }
    if (alive[i][nunits]) // bias
      net += (weight[i][nunits]);
				      
    netinput[i] = net;
    if (i<nunits-noutputs)// i in hidden schicht
    {
      activation[i] = (double) log_hidden(net);
      un_av[i]+=activation[i];
      un_var[i]+=pow(activation[i], 2);
      if (activation[i]<un_min[i]) un_min[i]=activation[i];
      if (activation[i]>un_max[i]) un_max[i]=activation[i];
      if (activation[i]>up_border) un_upper[i]++;
      if (activation[i]<low_border) un_lower[i]++;
    }
    else// i in output schicht
    {
      activation[i] = (double) log_output(net);
    }
  }
//  write_netinputs(); // zum Testen
}

void Netz::compute_error() {
  register int i,j;

  #ifdef DEBUG
  cout << "\nfms|Netz::compute_error()";
  #endif
  
  E=0.0;
  
  for (i = ninputs; i < nunits - noutputs; i++) {
    u_error[i] = 0.0;
  }

  // i==nunits-noutputs
  // Fehler nur fuer die Output-units berechnen
  // Rest wird spaeter in compute_wed() berechnet
  for (j = 0; i < nunits; j++, i++) {
    u_error[i] = target[j] - activation[i];
    E+=u_error[i]*u_error[i];  
  }
  tss+=E;
  E_n+=E;
  if (E>pss_max)
    pss_max = E;
  if (E<ecrit) correct_class++;
}

void Netz::compute_wed() {
  register int   k, i, j;
  int grenze1, grenze2;
  boolean k_i=FALSE;
  
  #ifdef DEBUG
  cout << "\nfms|Netz::compute_wed()";
  #endif

  
  for (k=nunits-noutputs; k<nunits; k++) //Schleife output-units
  {
    ks[k][k] = log_out_strich(netinput[k]);
    for (j=nunits-noutputs-1; j >=0; j--) 
    { // Schleife hidden-input-units
      if (j>=ninputs) // j in hidden schicht
      {
        grenze1=nunits-1;
        grenze2=nunits-noutputs;
      }
      else// j in input schicht
      {
        grenze1=nunits-noutputs-1;
        grenze2=ninputs;
      }
      
      for (i=grenze1; i>=grenze2;i--) 
      { // Schleife output-hidden-units
	if (j >= ninputs) 
        {
          i=k;
          k_i=TRUE;
        }
        // Zwischenwerte berechnen
        if (alive[i][j])
	{
	  ykw[k][i][j]=activation[j]*ks[k][i];
	  if (fabs(ykw[k][i][j]) < .000001) ykw[k][i][j]=sign(ykw[k][i][j])*.000001; // um Divisions-Ueberlauf zu vermeiden
	  yw[i][j]+=ykw[k][i][j]*u_error[k]/tpatterns; // zur Anpassung von alpha an die Anzzahl der Muster
 	  ky[k][j]+=weight[i][j]*ks[k][i];
	  t_eins[i][j]+=ykw[k][i][j]*ykw[k][i][j];
	}
        if (alive[i][nunits]&&((j==nunits-noutputs-1)||(j==0)))
        {
          ykw[k][i][nunits]=1.0*ks[k][i];
          if (fabs(ykw[k][i][nunits]) < .000001) ykw[k][i][nunits]=sign(ykw[k][i][nunits])*.000001;// um Divisions-Ueberlauf zu vermeiden
          yw[i][nunits]+=ykw[k][i][nunits]*u_error[k]/tpatterns; // zur Anpassung von alpha an die Anzzahl der Muster
          ky[k][nunits]+=weight[i][nunits]*ks[k][i];
          t_eins[i][nunits]+=ykw[k][i][nunits]*ykw[k][i][nunits];
        }
        // Zwischenwerte berechnen Ende
	if (k_i) { 
	  k_i=FALSE;
          break;
        }
      }
      if (j<ninputs) // j in input schicht
      {
	ks[k][j]=0.0; // Ableitung der Input-Units==0!
      }
      else// j in hidden schicht
      {  
        ks[k][j] = log_hid_strich(netinput[j])*ky[k][j];
      }
      ks[k][nunits]=0.0; // Ableitung nach netzinput der bias-unit
    }
  }
}

void Netz::compute_t2_t3()   // t2 und t3 berechnen 
{
  register int   k, i, j;
  int grenze1, grenze2;
  
  #ifdef DEBUG
  cout << "\nfms|Netz::compute_t2_t3()";
  #endif

  for (k=nunits-noutputs; k<nunits; k++) //Schleife output-units
  {
    for (j=nunits-noutputs-1; j >=0; j--) { // Schleife hidden-input-units
      if (j<ninputs)//input-schicht
      {
	grenze1=nunits-1;
	grenze2=nunits-noutputs;
      }
      else
      {
        grenze1=nunits-noutputs-1;
        grenze2=ninputs;
      }				  
      for (i=grenze1; i>=grenze2;i--) // Schleife output-hidden-units
      {
        if (alive[i][j])
	{
	  t_zwei[k]+=fabs(ykw[k][i][j])/sqrt(t_eins[i][j]);
	}
        if (alive[i][nunits]&&((j==0)||(j==nunits-noutputs-1)))
	{
	  t_zwei[k]+=fabs(ykw[k][i][nunits])/sqrt(t_eins[i][nunits]);
	}
      }
    }
    t_drei+=t_zwei[k]*t_zwei[k];
  }
}

void Netz::compute_insig_weights()   // insignifikante Gewichte berechnen 
{
  register int  k, i, j;
  int grenze1, grenze2;
  
  #ifdef DEBUG
  cout << "\nfms|Netz::compute_insig_weights()";
  #endif

  sig_weights = 0;
   
  // Gewichte auf 0 setzen
  for (i = ninputs; i < nunits; i++) {// hidden-output
    if (i<nunits-noutputs)//i in hidden schicht
    {
      grenze1=0;
      grenze2=ninputs;
    }
    else//i in output schicht
    {
      grenze1=ninputs;
      grenze2=nunits-noutputs;
    }
				      
    for (j=grenze1; j < grenze2; j++) {// input-hidden
      if (alive[i][j])
      {
	del_w[i][j]=sqrt(epsilon)/(sqrt(t_eins[i][j])*sqrt(t_drei));
	if (del_w[i][j]<del_min) del_min=del_w[i][j];   
        if (del_w[i][j]>del_max) del_max=del_w[i][j];
        B+= -log10(del_w[i][j]);
      }
    }
    // bias
    if (alive[i][nunits])
    {
      del_w[i][nunits]=sqrt(epsilon)/(sqrt(t_eins[i][nunits])*sqrt(t_drei));
      if (del_w[i][nunits]<del_min) del_min=del_w[i][nunits];   
      if (del_w[i][nunits]>del_max) del_max=del_w[i][nunits];
      B+= -log10(del_w[i][nunits]);
    }
  }
  
  for (i = ninputs; i < nunits; i++) {// hidden-output
    if (i<nunits-noutputs)//i in hidden schicht
    {
      grenze1=0;
      grenze2=ninputs;
    }
    else//i in output schicht
    {
      grenze1=ninputs;
      grenze2=nunits-noutputs;
    }

    for (j=0; j < nunits-noutputs; j++) {//input-hidden
      if (alive[i][j])
      {
        if (del_w[i][j]>K*del_min)
	{
          insignificant[i][j]=TRUE;
	  for (k=nunits-noutputs;k<nunits;k++)
	  {
	    t_eins[i][j]-=(ykw[k][i][j]*ykw[k][i][j]);
          }
	}
        else
        {
	  sig_weights++;
	  wasalive[i][j]=TRUE;
	}
      }
    }
    // bias
    if (alive[i][nunits])
    {
      if (del_w[i][nunits]>K*del_min)
      {
        insignificant[i][nunits]=TRUE;
        for (k=nunits-noutputs;k<nunits;k++)
        {
          t_eins[i][nunits]-=(ykw[k][i][nunits]*ykw[k][i][nunits]);
        }
      }
      else
      {
        sig_weights++;
        wasalive[i][nunits]=TRUE;
      }
    }
  }
  
  if (sig_weights<min_sigs)
    min_sigs=sig_weights;
  if (sig_weights>max_sigs)
    max_sigs=sig_weights;
  av_sigs+=sig_weights;
}

void Netz::update_variables()
{
  int k, i, j;
  int grenze1, grenze2;
  
  #ifdef DEBUG
  cout << "\nfms|Netz::update_variables()";
  #endif

  t_drei=0;
  for (k=nunits-noutputs;k<nunits;k++) 
  {
    t_zwei[k]=0;
    for (i = ninputs; i < nunits; i++) {// hidden-outputs
      if (i<nunits-noutputs)//i in hidden schicht
      {
	grenze1=0;
	grenze2=ninputs;
      }
      else//i in output schicht
      {
	grenze1=ninputs;
	grenze2=nunits-noutputs;
      }

      for (j=grenze1; j < grenze2; j++) {// input-hidden
        if (alive[i][j]&&!insignificant[i][j])
        {
	  t_zwei[k]+=fabs(ykw[k][i][j])/sqrt(t_eins[i][j]);
        }
      }
      // bias
      if (alive[i][nunits]&&!insignificant[i][nunits])
      {
        t_zwei[k]+=fabs(ykw[k][i][nunits])/sqrt(t_eins[i][nunits]);
      }
    }
    t_drei+=t_zwei[k]*t_zwei[k];
  }
}

void Netz::compute_t4()
{
  int k, m, i, j;
  int grenze1, grenze2;
  
  #ifdef DEBUG
  cout << "\nfms|Netz::compute_t4()";
  #endif

  for (k=nunits-noutputs;k<nunits;k++) 
  {
    for (i = ninputs; i < nunits; i++) {// hidden-output
      if (i<nunits-noutputs)//i in hidden schicht
      {
        grenze1=0;
        grenze2=ninputs;
      }
      else//i in output schicht
      {
        grenze1=ninputs;
        grenze2=nunits-noutputs;
      }

      for (j=grenze1; j < grenze2; j++) {// input-hidden
        if (alive[i][j]&&!insignificant[i][j])
        {
	  t_vier=0;
          for (m=nunits-noutputs;m<nunits;m++)
          {
	    t_vier+=t_zwei[m]*sign(ykw[m][i][j])*(kron(m,k)*t_eins[i][j]-ykw[m][i][j]*ykw[k][i][j])/sqrt(t_eins[i][j]*t_eins[i][j]*t_eins[i][j]);
          }	
          b[k][i][j]=ykw[k][i][j]/t_eins[i][j]+sig_weights*t_vier/t_drei;  
        }
      }
      // bias
      if (alive[i][nunits]&&!insignificant[i][nunits])
      {
        t_vier=0;
        for (m=nunits-noutputs;m<nunits;m++)
        {
          t_vier+=t_zwei[m]*sign(ykw[m][i][nunits])*(kron(m,k)*t_eins[i][nunits]-ykw[m][i][nunits]*ykw[k][i][nunits])/sqrt(t_eins[i][nunits]*t_eins[i][nunits]*t_eins[i][nunits]);
        }	
        b[k][i][nunits]=ykw[k][i][nunits]/t_eins[i][nunits]+sig_weights*t_vier/t_drei;  
      }
    }
    
  }
}


void Netz::sec_forward()       // zweiter forward-pass
{
  int k, i, j, grenze1, grenze2;
  boolean k_i=FALSE;
    
  #ifdef DEBUG
  cout << "\nfms|Netz::sec_forward()";
  #endif

  for (k=nunits-noutputs; k<nunits; k++) //Schleife output-units
  {
    for (j=0;j<ninputs;j++)
    {
      ry[k][j]=0.0;
    }
    ry[k][nunits]=0.0;// bias
    for (i=ninputs; i <nunits; i++) { // Schleife hidden-output-units
      if (i >= nunits-noutputs){
        i=k;
        k_i=TRUE;
      }
      if (i>=nunits-noutputs)//output-schicht
      {
        grenze1=ninputs;
        grenze2=nunits-noutputs;
      }
      else//hidden-schicht
      {
        grenze1=0;
        grenze2=ninputs;
      }
      
      for (j=grenze1; j<grenze2;j++) // Schleife input-hidden-units
      {
        // Zwischenwerte berechnen
        if ((alive[i][j]) && !(insignificant[i][j]))
	{
	  rs[k][i]+=weight[i][j]*ry[k][j]+b[k][i][j]*activation[j];
	}
        // Zwischenwerte berechnen Ende
      }
      // bias
      // Zwischenwerte berechnen
      if ((alive[i][nunits]) && !(insignificant[i][nunits]))
      {
	rs[k][i]+=weight[i][nunits]*ry[k][i]+b[k][i][nunits]*1.0;
      }
      // Zwischenwerte berechnen Ende
      if (i<nunits-noutputs)
      { // hidden-units
        ry[k][i]=rs[k][i]*log_hid_strich(netinput[i]);
      }
      else
      {
        ry[k][i]=rs[k][i]*log_out_strich(netinput[i]);
      }
      if (k_i){
	k_i=FALSE;
        break;
      }
    }
  }
}

void Netz::sec_backward()       // zweiter backward-pass
{
  int k, i, j;
  boolean k_i=FALSE;
  int grenze1, grenze2;
  
  #ifdef DEBUG
  cout << "\nfms|Netz::sec_backward()";
  #endif

  // bias
  rdky[nunits]=0;
  
  for (k=nunits-noutputs; k<nunits; k++) //Schleife output-units
  {
    rdks[k]=rs[k][k]*log_out_2strich(netinput[k]);
    for (j=nunits-noutputs-1; j >=0; j--)  // Schleife hidden-input-units
    {
      rdky[j]=0;
      if (j>=ninputs)//j in hidden schicht
      {
        grenze1=nunits-1;
        grenze2=nunits-noutputs;
      }
      else//j in output schicht
      {
        grenze1=nunits-noutputs-1;
        grenze2=ninputs;
      }

      for (i=grenze1; i>=grenze2;i--)
      // Schleife output-hidden-units
      {
	if (j >= ninputs) {
          i=k;
          k_i=TRUE;
        } 
        if (alive[i][j]&&!insignificant[i][j])
	{
	  rdky[j]+=weight[i][j]*rdks[i]+b[k][i][j]*ks[k][i];
	  rdw[i][j]+=activation[j]*rdks[i]+ry[k][j]*ks[k][i];
	}
        // bias
        if  ((alive[i][nunits])&&(!(insignificant[i][nunits]))&&((j==nunits-noutputs-1)||(j==0)))
	{
//	  rdky[nunits]+=weight[i][nunits]*rdks[i]+b[k][i][nunits]*ks[k][i];
  	  rdw[i][nunits]+=1.0*rdks[i]+ry[k][nunits]*ks[k][i];
	}

        if (j<ninputs)
        { //input-units
          rdks[j]=0;
        }        
        else 
	{ // hidden-units
          rdks[j]=log_hid_strich(netinput[j])*rdky[j]+rs[k][j]*ky[k][j]*log_hid_2strich(netinput[j]);
        }
	if (k_i){
	  k_i=FALSE;
          break;
        }
      }
    }
  }
}

void Netz::norm_grad_b()  // Normalisierung des Gradienten von B
{
  int c, i, j, grenze1, grenze2;
    
  #ifdef DEBUG
  cout << "\nfms|Netz::norm_grad_b()";
  #endif
  
  c=0;
  lyw=lrdw=0.0;
  for (i = ninputs; i < nunits; i++) {// hidden-output
    if (i<nunits-noutputs)//i in hidden schicht
    {
      grenze1=0;
      grenze2=ninputs;
    }
    else//i in output schicht
    {
      grenze1=ninputs;
      grenze2=nunits-noutputs;
    }
    for (j=grenze1; j < grenze2; j++) {// input-hidden
      if (alive[i][j]&&!insignificant[i][j])
      {
        lyw+=yw[i][j]*yw[i][j];
        lrdw+=rdw[i][j]*rdw[i][j];
        c++;
      }
    }
    // bias
    if (alive[i][nunits]&&!insignificant[i][nunits])
    {
      lyw+=yw[i][nunits]*yw[i][nunits];
      lrdw+=rdw[i][nunits]*rdw[i][nunits];
      c++;
    }
  } 
  // wird unendlich, wenn alle Gewichte insignifikant!!
  scale=sqrt(lyw)/sqrt(lrdw); // scale=||yw||/||rdw||
  if(isnan(scale))
    scale = 0;
}

void Netz::change_weights() {
  register int i, j;
  int grenze1, grenze2;
  
  #ifdef DEBUG
  cout << "\nfms|Netz::change_weights()";
  #endif
  for (i = ninputs; i < nunits; i++) { // hidden-output
    if (i<nunits-noutputs)//hidden-schicht
    {
      grenze1=0;
      grenze2=ninputs;
    }
    else      
    {
      grenze1=ninputs;
      grenze2=nunits-noutputs;
    }

    for (j=grenze1; j < grenze2; j++) { // input-hidden
      if (alive[i][j]&&!insignificant[i][j])
      {
	weight[i][j]+=alpha*yw[i][j]-scale*lambda*rdw[i][j];
      }
    }

    // bias
    if (alive[i][nunits]&&!insignificant[i][nunits])
    {
      weight[i][nunits]+=balpha*yw[i][nunits]-scale*lambda*rdw[i][nunits];
    }
  } 
}

void Netz::setinput() {
  register int i;

  #ifdef DEBUG
  cout << "\nfms|Netz::setinput()";
  #endif

  for (i = 0; i < ninputs; i++) {
    activation[i] = ipattern[patno][i];
  }
  strcpy(cpname,pname[patno]);
}

void Netz::settarget() {
  register int k;

  #ifdef DEBUG
  cout << "\nfms|Netz::settarget()";
  #endif
  for (k = 0; k < noutputs; k++) {
    target[k] = tpattern[patno][k];
  }
}

void Netz::setup_pattern() {
  #ifdef DEBUG
  cout << "\nfms|Netz::setup_pattern()";
  #endif
  setinput();
  settarget();
}

int Netz::trial() {
  #ifdef DEBUG
  cout << "\nfms|Netz::trial()";
  #endif
  
  setup_pattern();

  compute_output();
  compute_error();
  return (CONTINUE);
}

int Netz::ptrain() {
  #ifdef DEBUG
  cout << "\nfms|Netz::ptrain()";
  #endif
  return(train('p'));
}

int Netz::strain() {
  #ifdef DEBUG
  cout << "\nfms|Netz::strain()";
  #endif
  return(train('s'));
}

int Netz::train(char c) 
{
  int t,j,i,old,npat, part;
  int k, grenze1, grenze2;
  char *str;
  ofstream gewichte;
  double E_a_old, E_n_old;
  
  str=NULL;
  #ifdef DEBUG
  cout << "\nfms|Netz::train()";
  #endif
  if (!System_Defined)
      return(BREAK);

  
  // Initialisierung
  E_a = E_n = tss = 0.0;
  E_o = pow(10, 6);
  // Ende Initialisierung

  // Anfang Lernen
  for (t = 0; (t < nepochs) || ((t>=nepochs)&&(E_n >= ecrit)); t++) {
    // er soll solange lernen, bis t>=nepochs und dann auch noch E_n unter ecrit ist
    if (lflag) epochno++; // Epoche hochzaehlen
    // Musterreihenfolge initialisieren
    for (i = 0; i < npatterns; i++) used[i] = i;
    if (choose_string[0] == 'r') {
    // Reihenfolge der Muster zufaellig festlegen
      for (i = 0; i < npatterns; i++) {
        npat = ((int)(rnd()*npatterns) % (npatterns-i)) + i;
        old = used[i];
        used[i] = used[npat];
        used[npat] = old;
      }
    }
    // Gewichtsupdate nach tpatterns Muster
    av_sigs=0;
    min_sigs=nunits*nunits+nunits;//==Anzahl der Gewichte
    max_sigs=0;
    for (k=0;k<nunits;k++)
    {
      un_av[k]=un_var[k]=un_max[k]=0.0;
      un_min[k]=1.0;
      un_upper[k]=un_lower[k]=0;
    }
    for (part=0; part<(npatterns/tpatterns); part++)
    {
	
    for (k=0;k<nunits;k++)
    {
      for (j=0;j<nunits;j++) 
      {
        yw[k][j]=0.0;
        rdw[k][j]=0.0;
      }
      // bias
      yw[k][nunits]=0.0;
      rdw[k][nunits]=0.0;
    }
    // Epochendurchlauf: alle Muster einmal trainieren
    for (i = part*tpatterns; i < (part+1)*tpatterns; i++) {
      // Initialisierung
      del_min=10;  // del_min auf hohen Wert setzen
      del_max=0;   // del_max auf 0 setzen
      // netinput[], yw[][], ky[][], t_eins[][], t_zwei[][], rs[][], rdw[][], rdky[][] auf 0 setzen
      for (k=0;k<nunits;k++)
      {
	for (j=0;j<nunits;j++) 
	{
	  t_eins[k][j]=0.0;
          // nur fuer k in output-units
          ky[k][j]=0.0;
          rs[k][j]=0.0;
        }
        // bias
	t_eins[k][nunits]=0.0;
        ky[k][nunits]=0.0;
        rs[k][nunits]=0.0;
        // restl. k
        netinput[k] = 0.0;
        rdky[k]=0.0;
        t_zwei[k]=0;
      }
      t_drei=t_vier = 0.0;
      for (k = ninputs; k < nunits; k++) {// hidden-output
        if (k<nunits-noutputs)
        {
	  grenze1=0;
	  grenze2=ninputs;
	}
	else
        {
	  grenze1=ninputs;
	  grenze2=nunits-noutputs;
        }				  
        for (j =grenze1; j < grenze2; j++){// input-hidden
          insignificant[k][j] = FALSE;
        }
        // bias
        if (k>=ninputs)
          insignificant[k][nunits]=FALSE;
      }

      // Ende Initialisierung
      patno = used[i];
      // in trial: setuppattern(), compute_output(), compute_error()
      if (trial() == BREAK) return (BREAK);
      #ifdef DEBUG
      cout << "\nMuster: " << i << " Epoche: " << epochno+old_epochs;
      #endif

      // Lern-Modus
      if (lflag) {
        compute_wed();
        // FMS **********************
	compute_t2_t3();
        compute_insig_weights();
        update_variables();
        compute_t4();
        sec_forward();
        sec_backward();
        norm_grad_b();
       }
      // nach jedem Muster Anzeige updaten
      if (step_size == PATTERN) {
        #ifndef DEBUG
        update_display(); 
        #endif
      }
      exemplars+=1;
    }
    // *********************************
    if (lflag && (grain_string[0] == 'e'))
    {
      change_weights();
    }

    }// Ende for part...
    
    av_sigs/=npatterns;
    E_n=E_n/npatterns;  
    B=B/npatterns;
    if (init_error==FALSE) { 
      E_a=E_n;// damit E_a richtiger initialisiert wird
      B_a=B;
      init_error=TRUE;
    }
    E_a=E_gamma*E_a+(1-E_gamma)*E_n;
    B_a=E_gamma*B_a+(1-E_gamma)*B;
    for (k=0; k<nunits; k++)
    {
      un_av[k]/=npatterns;// durchschnittliche Aktivierung einer Unit
    }

    char fname[BUFSIZ];
    strcpy(fname, net_name);
    strcat(fname, ".log");  
    gewichte.open(fname, ios::app);
    gewichte.width(4);
    gewichte.precision(6);
    gewichte.setf(ios::fixed);
    if (lflag)
      gewichte << "\nEp. " << epochno << "\tE_n/E_a/B/B_a/av_sigs/min_sigs/max_sigs " << E_n << "\t" << E_a << "\t" << B << "\t" << B_a << "\t " << av_sigs << "\t" << min_sigs << "\t" << max_sigs;
    else
      gewichte << "\nTestEp.\tE_n/B/av_sigs/min_sigs/max_sigs " << E_n << "\t" << B << "\t " << av_sigs << "\t" << min_sigs << "\t" << max_sigs;

    if ((E_o<E_n)&&lflag)
    {
      // Ausgabe in logdatei wenn E_n zugenommen
      gewichte << " E-gestiegen E_o: " << E_o;
    }
    gewichte.close();
    
    comp_alive_units();
    // nach jeder Epoche Anzeige updaten
    if (step_size == EPOCH) {
      #ifndef DEBUG
      update_display();
      #endif
    }
    if ((((epochno+old_epochs) % lparam_update) == 0)&&lflag) {
      update_lparam();
      write_activations();
      write_alives();
    }
    // Werte ruecksetzen
    E_o=E_n;
    E_n=B=tss=pss_max=0.0;
    correct_class=0;
    // ***************************
    if ((((epochno+old_epochs) % lparam_update) == 0)&&lflag) {
      write_weights();
      write_unit_acts();
      if ((epochno+old_epochs)%(lparam_update*10) == 0)
      {
        E_a_old=E_a;
        E_n_old=E_n;
        test_all();
        E_a=E_a_old;
        E_n=E_n_old;
      }
    }
  } // Epochen-Ende/ Lern-Ende
  
  // nach allen Epochen Anzeige updaten
  if (step_size == NEPOCHS) {
    #ifndef DEBUG     
    update_display();
    #endif
  }
  return(CONTINUE);
}

int Netz::test_all() {
  // zum Testen aller Muster
  int save_single_flag;
  int save_nepochs;
  int save_step_size;
   
  #ifdef DEBUG
  cout << "\nfms|Netz::tall()";
  #endif
  lflag = 0;
  save_single_flag = single_flag; 
  save_step_size = step_size; 
  if (step_size > PATTERN) step_size = PATTERN;
  save_nepochs = nepochs;  
  nepochs = 1;
  if (PAIRS)
    get_pattern_pairs();
  else
    get_patterns();
  train('s');
  nepochs = save_nepochs;
  single_flag = save_single_flag;
  step_size = save_step_size;
  lflag=1;
  if (PAIRS)
    get_pattern_pairs();
  else
    get_patterns();
  return(CONTINUE);
}
  
void Netz::init_params()      // Parameter initialisieren
{
  int i,j,k;
    
  // Parameter initialisieren, falls keine Werte in .net gesetzt werden
  #ifdef DEBUG
  cout << "\nfms|Netz::init_params()";
  #endif

  alive = new boolean *[nunits];
  wasalive = new boolean *[nunits];
  insignificant = new boolean *[nunits];
  t_eins = new double *[nunits];
  del_w = new double *[nunits]; 
  yw = new double *[nunits]; 
  un_av = new double[nunits];
  un_var = new double[nunits];
  un_min = new double[nunits];
  un_max = new double[nunits];
  un_upper = new int[nunits];
  un_lower = new int[nunits];
  treated =new boolean[nunits];
  rdw = new double *[nunits]; 
  rdks = new double [nunits+1]; 
  rdky = new double [nunits+1]; 
  for (i = 0; i < nunits; i++) {
    alive[i] = new boolean[nunits+1];
    wasalive[i] = new boolean[nunits+1];
    insignificant[i] = new boolean[nunits+1];
    t_eins[i]= new double[nunits+1];
    del_w[i]= new double[nunits+1];
    yw[i] = new double[nunits+1];
    un_av[i]=un_var[i]=un_max[i]=0.0;
    un_min[i]=1.0;
    un_lower[i]=un_upper[i]=0;
    treated[i]=FALSE;
    rdw[i]= new double[nunits+1];
    rdks[i]= 0.0;
    rdky[i]= 0.0;
    for (j = 0; j < nunits+1; j++){ // +1 wg. bias!
      alive[i][j] = FALSE;
      wasalive[i][j] = FALSE;
      insignificant[i][j] = TRUE;
      t_eins[i][j]=0.0;
      del_w[i][j]=0.0;
      yw[i][j]=0.0;
      rdw[i][j]=0.0;
    }
  }
  rdks[nunits]= 0.0;
  rdky[nunits]= 0.0;
  
  // alive[][] TRUE setzen, fuer alle Gewichte, die tatsaechlich vorkommen
  for (i=0; i< nunits-noutputs; i++)// input-hidden
  {
    for (j=ninputs; j< nunits; j++)// hidden-output
    {
      if (((i<ninputs)&&(j<nunits-noutputs))||((i>=ninputs)&&(j>=nunits-noutputs)))
        alive[j][i]=TRUE;
      else
      {
        weight[j][i]=0.0;
        alive[j][i]=FALSE;
      }
      
      alive[j][nunits]=TRUE; //bias
    }
  }
  
  t_zwei=new double[nunits];
  
  ky=new double *[nunits];
  ks=new double *[nunits];
  ry=new double *[nunits];
  rs=new double *[nunits];
  ykw=new double **[nunits];
  b=new double **[nunits];
  for (k=0;k<nunits;k++)
  {  
    t_zwei[k]=0;
      
    ky[k]=new double [nunits+1];
    ks[k]=new double [nunits+1];
    ry[k]=new double [nunits+1];
    rs[k]=new double [nunits];
    ykw[k]=new double *[nunits];
    b[k]=new double *[nunits];
    
    for (j=0;j<nunits;j++)
    {
      ky[k][j]=0.0;
      ks[k][j]=0.0;
      ry[k][j]=0.0;
      rs[k][j]=0.0;
      ykw[k][j]=new double [nunits+1];
      b[k][j]=new double [nunits+1];
      for (i=0;i<nunits+1;i++)
      {
        ykw[k][j][i]=0.0;
        b[k][j][i]=0.0;
      } 
    }
    ky[k][nunits]=0.0;
    ks[k][nunits]=0.0;
    ry[k][nunits]=0.0;
  }
  
  del_min=10; // del_min auf hohen Wert setzen
  del_max=0;       // del_max auf 0 setzen
  // Werte aus Datei lesen
  read_alives();
}

void Netz::init_weights() {
  #ifdef DEBUG
  cout << "\nfms|Netz::init_weights()";
  #endif
  // Gewichte einlesen, falls Datei vorhanden, sonst gemaess Angaben initialisieren
  read_weights();
}

void Netz::update_lparam() {
  int i, j, m, grenze1, grenze2;
  int grenze=0;
  char fname[BUFSIZ];
  ofstream logdatei;
  
  #ifdef DEBUG
  cout << "\nfms|Netz::update_lparam()";
  #endif

  are_alive=0;

  // lambda (2. Lernrate) anpassen
  if ((E_n<ecrit) || (E_n<=E_o)) lambda+=delta_lambda;
  else 
  if (E_n<=E_a)
  {
    lambda-=delta_lambda;
    if (lambda<0) lambda=0;
  }
  else
    lambda*=sigma;

  strcpy(fname, net_name);
  strcat(fname, ".log");  
  logdatei.open(fname, ios::app);
  logdatei << "\nEp. " << epochno << "\tlambda/alpha:\t\t" << lambda << "\t" << alpha;
  logdatei.close();
  
  // alive und wasalive anpassen
  for (i = ninputs; i < nunits; i++) {// hidden-output
    if (i>=nunits-noutputs) // i in output schicht
    {
      grenze1=ninputs;
      grenze2=nunits-noutputs;
    }
    else// i in hidden schicht
    {
      grenze1=0;
      grenze2=ninputs;
    }
    
    for (j=grenze1; j<grenze2; j++)// input-hidden
    {
      if ((alive[i][j])) 
      {
        if ((wasalive[i][j]==FALSE)&&(grenze<=noutputs)) {// nur eine Unit darf pro Durchlauf geloescht werden
	  grenze++;
          alive[i][j]=FALSE;
          // Ausgabe in logdatei
          strcpy(fname, net_name);
          strcat(fname, ".log");  
          logdatei.open(fname, ios::app);
          logdatei << "\nEp. " << epochno << "\t[" << i << "][" << j << "] auf nicht alive";
          logdatei.close();
        }
        wasalive[i][j]=FALSE;
        if (alive[i][j]) are_alive+=1;
      }
    } // for j...
    // bias
    if (alive[i][nunits]) 
    {
      if ((wasalive[i][nunits]==FALSE)&&(grenze<=noutputs)) {
        grenze++;
	{
          alive[i][nunits]=FALSE;
          // Gewichte in das darueberliegende Layer verteilen
          if (i<nunits-noutputs) // ist in hidden-schicht
            for (m=nunits-noutputs; m<nunits; m++)
	      if (alive[m][i])
	      {
                // bias verteilen
	      }
          // Ausgabe in logdatei
          strcpy(fname, net_name);
          strcat(fname, ".log");  
          logdatei.open(fname, ios::app);
          logdatei << "\nEpoche: " << epochno << "\tbias[" << i << "] auf nicht alive";
          logdatei.close();
        }
      }
      wasalive[i][nunits]=FALSE;
      if (alive[i][nunits]) are_alive+=1;
    }
  }
//  if ((epochno % 100 == 0)||(E_a>2.0*ecrit))
/*  if ((epochno % 100 == 0))
  {
    // alive[][] TRUE setzen, fuer alle Gewichte, die tatsaechlich vorkommen
    for (i=0; i< nunits-noutputs; i++)// input-hidden
    {
      if (i<ninputs) //i in input schicht
      {
	grenze1=ninputs;
        grenze2=nunits-noutputs;
      }    
      else// i in hidden schicht
      {
	grenze1=nunits-noutputs;
        grenze2=nunits;
      }
      for (j=grenze1; j< grenze2; j++)// hidden-output
      {
	alive[j][i]=TRUE;
        alive[j][nunits]=TRUE;
      }
    }
  }
*/
}

int Netz::write_activations() {
  int     i;
  char fname[BUFSIZ];
  ofstream act_file;
  boolean datei_ok = FALSE;
  
  #ifdef DEBUG
  cout << "\nfms|Netz::write_activations()";
  #endif

  do
  {
    if (!nodisplay) 
    {
      strcpy(fname, net_name);
  
      strcat(fname, ".act");  
      if (fname == NULL) 
        return(CONTINUE);
    }
    else {
      strcpy(fname, net_name);
  
      strcat(fname, ".act");  
      if (fname == NULL) 
        return(CONTINUE);
    }  
   
    act_file.open(fname);
    if (act_file.fail()) {
      put_error("\nFehler beim Oeffnen der Aktivierungsdatei.");
      return(BREAK); // Datei nicht vorhanden
    }
    datei_ok=TRUE;
  }
  while (!datei_ok);

  act_file << "Epoche: " << epochno+old_epochs << "\n";
  act_file << "activation: \n";
  for (i = 0; i < nunits; i++) {
    #ifdef DEBUG
    cout << "\nSchreibe Aktivierung [" << i << "]" << activation[i] << "\n";
    #endif
    act_file << activation[i] << " ";
    if ((i==ninputs-1)||(i==nunits-noutputs-1)) act_file << "\n";
  }

  return(CONTINUE);
}

int Netz::write_alives()       // schreibt aktuelle alive[][], wasalive[][] xxxxxxxx.alive
{
  int i, j;
  char fname[BUFSIZ];
  ofstream alive_file;
  boolean datei_ok = FALSE;
  
  #ifdef DEBUG
  cout << "\nfms|Netz::write_alives()";
  #endif

  do
  {
    if (!nodisplay) 
    {
      strcpy(fname, net_name);
  
      strcat(fname, ".alive");  
      if (fname == NULL) 
        return(CONTINUE);
    }
    else {
      strcpy(fname, net_name);
  
      strcat(fname, ".alive");  
      if (fname == NULL) 
        return(CONTINUE);
    }  
   
    alive_file.open(fname);
    if (alive_file.fail()) {
      put_error("\nFehler beim Oeffnen der alive-datei.");
      return(BREAK); // Datei nicht vorhanden
    }
    datei_ok=TRUE;
  }
  while (!datei_ok);

  for (i = 0; i < nunits; i++) {
    for (j=0; j<nunits+1; j++)
    {
      #ifdef DEBUG
      cout << "\nSchreibe alive und wasalive[" << i << "]["<< j <<"] " << activation[i] << "\n";
      #endif
      if (alive[i][j]) 
        alive_file << "1";
      else
        alive_file << "0";
    }
    alive_file << "\n";
    if ((i==ninputs-1)||(i==nunits-noutputs-1)) alive_file << "\n";
  }
  alive_file << "\n";
  
  for (i = 0; i < nunits; i++) {
    for (j=0; j<nunits+1; j++)
    {
      if (wasalive[i][j]) 
        alive_file << "1";
      else
        alive_file << "0";
    }
    alive_file << "\n";
    if ((i==ninputs-1)||(i==nunits-noutputs-1)) alive_file << "\n";
  }
  alive_file << "Epoche: " << epochno+old_epochs << " alive: "<< are_alive<< "\n";
  alive_file.close();
  return(CONTINUE);
}

int Netz::read_alives() 
{
  int i, j;
  char fname[BUFSIZ];
  ifstream alive_file;
  char tempss[nunits+10];
  if (!nodisplay) 
  {
    strcpy(fname, net_name);

    strcat(fname, ".alive");  
    if (fname == NULL) 
      return(CONTINUE);
  }
  else {
    strcpy(fname, net_name);

    strcat(fname, ".alive");  
    if (fname == NULL) 
      return(CONTINUE);
  }  
   
  alive_file.open(fname);
  if (alive_file.fail()) {
    put_error("\nFehler beim Oeffnen der Alivedatei.(Datei ist nicht vorhanden?)");
    return(BREAK); // Datei nicht vorhanden
  }

  for (i = 0; i < nunits; i++) {
      alive_file >> tempss;
    for (j=0;j<nunits+1; j++) 
    {
      if (tempss[j]=='1') alive[i][j]=TRUE;
      else alive[i][j]=FALSE;
      
      if (alive_file.eof()) {
        return(put_error("Alivedatei ist nicht korrekt fuer dieses Netzwerk. Beim Lesen der Alive-Daten."));
      }
    }
  }
  for (i = 0; i < nunits; i++) {
      alive_file >> tempss;
    for (j=0;j<nunits+1; j++) 
    {
      if (tempss[j]=='1') wasalive[i][j]=TRUE;
      else wasalive[i][j]=FALSE;
      if (alive_file.eof()) {
        return(put_error("Alivedatei ist nicht korrekt fuer dieses Netzwerk. Beim Lesen der Alive-Daten."));
      }
    }
  }
  return(CONTINUE);
}

int Netz::comp_alive_units()  // zaehlt alle Units, die noch nicht weggeloescht wurden
{
  boolean ist_alive=FALSE;
  ofstream gewichte;
  int i, j, grenze1, grenze2;
  alive_units=0;
  // unit ist nicht alive, wenn alle wegfuehrenden Gewichte nicht alive sind oder wenn alle hinfuehrenden Gewichte und der Bias nicht alive sind
  // wie ist es, wenn die durchschnittliche Aktivierung bei 0/1 liegt??
  // dabei interessieren nur die hidden-units
  char fname[BUFSIZ];
  strcpy(fname, net_name);
  strcat(fname, ".log");  
  gewichte.open(fname, ios::app);
  gewichte.width(4);
  gewichte.precision(6);
  gewichte.setf(ios::fixed);
  gewichte << "\nEp. " << epochno << "\talive-units: ";
  for (i=ninputs; i<nunits-noutputs; i++) {//hidden
    ist_alive=FALSE;
    if (!treated[i])
    {
      // hinfuehrende Gewichte ********************************************
      // i in hidden schicht
      {
        grenze1=0;
        grenze2=ninputs;
      }
      for (j=grenze1; j<grenze2; j++)//input-hidden
      {
        if(alive[i][j])
        {
	  ist_alive=TRUE;
          break;
        }
      }
      // bias: hier muss ist_alive nochFALSE sein
      if (!ist_alive)
      {
        if (un_av[i]>=up_border)
        {
          alive[i][nunits]=FALSE;
        }
	for (j=nunits-noutputs; j<nunits; j++)
        {
	    
          if(alive[j][i])
          {
            alive[j][i]=FALSE;
            // die wegfuehrenden Gewichte werden mit Gewichtung unit_average auf den bias der zielunit addiert
            weight[j][nunits]+=weight[i][nunits]*un_av[i];
          }
        }
      }
      if (!ist_alive) // wegfuehrende Gewichte auf not alive setzen
      {
	treated[i]=TRUE;
        for (j=nunits-noutputs; j<nunits; j++)
        {
          if(alive[j][i])
          {
            alive[j][i]=FALSE;
            // die wegfuehrenden Gewichte werden mit Gewichtung unit_average auf den bias der zielunit addiert 
            weight[j][nunits]+=weight[j][i]*un_av[i];
          }
        }
      }
    
      if (!ist_alive) continue;
     
      // wegfuehrende Gewichte: Unit ist noch alive **********************
      for (j=nunits-noutputs; j<nunits; j++)
      {
        // wegfuehrende Gewichte muessen not alive oder < .001 sein dann ist die unit weggeloescht!
        if(alive[j][i]&&(fabs(weight[j][i])>.001))
        {
          ist_alive=TRUE;
          break;
        }
        else
        if (alive[i][j]&&(fabs(weight[j][i])<=.001))
          alive[i][j]=FALSE;
      }
      if (!ist_alive)
      {
	treated[i]=TRUE;
        // hinfuehrende Gewichte auf not alive setzen
        for (j=0; j<ninputs; j++) 
          if (alive[i][j])
            alive[i][j]=FALSE;
        // bias auf not alive setzen
        if (alive[i][nunits]) alive[i][nunits]=FALSE;
        // wegfuehrende Gewichte auf not alive setzen     
        for (j=nunits-noutputs; j<nunits; j++) 
          if (alive[j][i])
            alive[j][i]=FALSE;
        }  
    }
    if (ist_alive) {
      alive_units++;
      gewichte << " " << i;
    }
  }
  gewichte << "\nEp. " << epochno << "\thidden-alive " << alive_units;
  gewichte.close();
  
  return(alive_units);
}

int Netz::update_display() {
  int i, j;
  ofstream ers_display;
  char   *str=new char [13];
  
  if (!lflag) return(CONTINUE);
  
  if (!nodisplay) 
  {
  
  #ifdef DEBUG
  cout << "\ndisplay|update_display()";
  #endif

  // in die erste Zeile erste Spalte gehen
  io_move(0,0);
  // summierten quadrierten Fehler anzeigen
  cout.width(4);
  cout.precision(6);
  cout.setf(ios::fixed);
  
  cout << "\ntss: " << tss << "\tE_n: " << E_n << "\tE_o: " << E_o << "\tE_a: " << E_a << "\nE_max: " << pss_max;
  cout << "\tlambda: " << lambda << "\talpha: " << alpha;
  cout << "\n";
  // Epoche anzeigen
  cout << "Epoche: " << epochno+old_epochs << "\tMuster: " << npatterns << "\tK: " << K << "\n";
  cout << "B: " << B << "\tB_a: " << B_a <<"\nsig_weights: " << sig_weights << "\tav_sigs: " << av_sigs << "\tmin_sigs: " << min_sigs << "\tmax_sigs: " << max_sigs << "\nalive: " << are_alive << "\thidden alive: " << alive_units;
  cout << "\n";
  cout << "del_min: " << del_min << "\tdel_max: " << del_max << "\n";
  cout << "Korrekt: " << correct_class << "\n";
  
  cout.setf(ios::fixed);
  cout.unsetf(ios::scientific);
  // Input-Units anzeigen
  // Zeile mit je 5 Units
  for (i=0;i<(ninputs/5.0);i++) {
    io_move(4+i, 1);
    if (ninputs-i*5>=5){
      for (j=0;j<5;j++)
      { cout.width(4);
        cout << (int)(activation[i*5+j]*1000) << " ";
      }
    }
    else {
      for (j=0;j<ninputs-(i*5);j++)
      { cout.width(4);
        cout << (int)(activation[i*5+j]*1000) << " ";
      }
    }
    cout << "\n";
  }
  // Hidden-Units anzeigen
  io_move(10,1);
  for (i=0;i<(nunits-ninputs-noutputs)/5.0;i++) {
    io_move(10+i,1);
    if ((nunits-ninputs-noutputs)-i*5>=5){
      for (j=0;j<5;j++)
      { cout.width(4);
        cout << (int)(activation[i*5+ninputs+j]*1000) << " ";
      }
    }
    else {
      for (j=0;j<(nunits-ninputs-noutputs)-i*5;j++)
      { cout.width(4);
        cout << (int)(activation[i*5+ninputs+j]*1000) << " ";
      }
    }
    cout << "\n";
  }

  // Output-Units anzeigen
  io_move(16, 1);
  for (i=0;i<noutputs/5.0;i++) {
    io_move(16+i, 1);
    if (noutputs-i*5>=5) {
      for (j=0;j<5;j++)
      { cout.width(4);
        cout << (int)(activation[nunits-noutputs+i*5+j]*1000) << " ";
      }
    }
    else {
      for (j=0;j<noutputs-5*i;j++)
      { cout.width(4);
        cout << (int)(activation[nunits-noutputs+i*5+j]*1000) << " ";
      }
    }
    cout << "\n";
  }

  } // if !nodisplay...
//  else wg. old_epochs!!!
  {
    strcpy(str, net_name);
    strcat(str, ".dsp");
    ers_display.open(str);
    if (ers_display.fail()) 
    {
      cout << "fehler bei stream oeffnen.";
    }
    ers_display.width(4);
    ers_display.precision(6);
    ers_display.setf(ios::fixed);
  
    ers_display << "tss: " << tss << "\tE_n: " << E_n << "\tE_o: " << E_o << "\tE_a: " << E_a << "\nE_max: " << pss_max;
    ers_display << "\tlambda: " << lambda << "\talpha: " << alpha;
    ers_display << "\n";
    // Epoche anzeigen
    ers_display << "Epoche: " << epochno+old_epochs << "\tMuster: " << npatterns << "\tK: " << K << "\n";
    ers_display << "B: " << B << "\tB_a: " << B_a <<"\nsig_weights: " << sig_weights << "\tav_sigs: " << av_sigs << "\tmin_sigs: " << min_sigs << "\tmax_sigs: " << max_sigs  << "\nalive: " << are_alive << "\thidden alive: " << alive_units;
    ers_display << "\n";
    ers_display << "del_min: " << del_min << "\tdel_max: " << del_max << "\n";
    ers_display << "Korrekt: " << correct_class << "\n";
    ers_display.close();
  }
  return(CONTINUE);
}

int Netz::write_netinputs() {
  int     i;
  char fname[BUFSIZ];
  ofstream netinput_file;
  boolean datei_ok = FALSE;
  

  #ifdef DEBUG
  cout << "\nNetz::write_netinputs()";
  #endif

  do
  {
    strcpy(fname, net_name);
  
    strcat(fname, ".netinput");
    if (fname == NULL) 
      return(CONTINUE);
   
    netinput_file.open(fname);
    if (netinput_file.fail()) {
      put_error("\nFehler beim Oeffnen der Netinputdatei.");
      return(BREAK); // Datei nicht vorhanden
    }
    datei_ok=TRUE;
  }
  while (!datei_ok);
  
  netinput_file << epochno+old_epochs << "\n";

  for (i = 0; i < nunits; i++) {
    #ifdef DEBUG
    cout << "netinput[" << i << "] " << netinput[i];
    #endif
    netinput_file << "netinput[" << i << "] " << netinput[i] << "\tactivation: " << activation[i] << "\tbias: "<< weight[i][nunits] <<"\n";
  }

  netinput_file.close();
  return(CONTINUE);
}

int Netz::write_unit_acts() // schreibt aktuelle un_av[], un_min[], un_max[], un_low[], un_up[] in x.units_epochno
{
  int i;
  char fname[BUFSIZ];
  ofstream alive_file;
  boolean datei_ok = FALSE;
  
  #ifdef DEBUG
  cout << "\nbp_enh|Netz::write_unit_acts()";
  #endif

  
  do
  {
    strcpy(fname, net_name);
  
    strcat(fname, ".units");  
    if (fname == NULL) 
      return(CONTINUE);

   
    alive_file.open(fname);
    if (alive_file.fail()) {
      put_error("\nFehler beim Oeffnen der alive-datei.");
      return(BREAK); // Datei nicht vorhanden
    }
    datei_ok=TRUE;
  }
  while (!datei_ok);
  alive_file << "un_av[i] un_min[i] un_max[i] un_lower[i] un_upper[i] un_var[i]";
  alive_file << "\n";

  for (i = ninputs; i < nunits-noutputs; i++) { //nur hidden-schicht
    alive_file << un_av[i] << " " << un_min[i] << " " << un_max[i] << " " << un_lower[i] << " " << un_upper[i] << " " << un_var[i];
    alive_file << "\n";
  }
  
  alive_file << "Epoche: " << epochno+old_epochs;
  alive_file.close();
  return(CONTINUE);
}

main()
{
  set_new_handler(heap_full);
  Netz MyNetz;
  
  
  MyNetz.Run(); 
} 


